/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsArchive.java 2530 2008-02-27 17:58:05Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.ant.archive;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;
import org.apache.tools.ant.types.ZipFileSet;
import org.ow2.easybeans.ant.archive.info.ArchiveInfo;

/**
 * Common task that creates archives.
 * @author Florent Benoit
 */
public class AbsArchive extends Task {

    /**
     * Reference to the standard deployment descriptor.
     */
    private File deploymentDescriptor = null;

    /**
     * Reference to the specific deployment descriptor.
     */
    private File specificDeploymentDescriptor = null;

    /**
     * Full path to the archive.
     */
    private File dest = null;

    /**
     * Relative Name of the archive (when embedded).
     */
    private String name = null;

    /**
     * Exploded mode or not ? (default = file).
     */
    private boolean exploded = false;

    /**
     * List of fileset used to add files.
     */
    private List<FileSet> fileSetList = null;

    /**
     * Default constructor.
     */
    public AbsArchive() {
        fileSetList = new ArrayList<FileSet>();
    }

    /**
     * Add the given fileset to the list of existing fileset.
     * @param zipFileSet the fileset to add.
     */
    public void addFileSet(final ZipFileSet zipFileSet) {
        fileSetList.add(zipFileSet);
    }

    /**
     * Gets the list of fileset to include in the archive.
     * @return the list of fileset to include in the archive.
     */
    public List<FileSet> getFileSetList() {
        return fileSetList;
    }

    /**
     * Sets the exploded mode to true or false.
     * @param exploded boolean true/false
     */
    public void setExploded(final boolean exploded) {
        this.exploded = exploded;
    }

    /**
     * Gets the state : exploded mode or not ?
     * @return the state : exploded mode or not ?
     */
    public boolean isExploded() {
        return exploded;
    }

    /**
     * Sets the reference to the deployment descriptor.
     * @param dd the given deployment descriptor.
     */
    public void setDD(final File dd) {
        // Ignore empty value
        if ("empty-value".equals(dd.getName())) {
            return;
        }

        if (!dd.exists()) {
            throw new BuildException("The given file '" + dd + "' for the deployment descriptor does not exist.");
        }
        this.deploymentDescriptor = dd;
    }

    /**
     * Sets the reference to the specific deployment descriptor.
     * @param dd the given specific deployment descriptor.
     */
    public void setSpecificDD(final File dd) {
        // Ignore empty value
        if ("empty-value".equals(dd.getName())) {
            return;
        }

        if (!dd.exists()) {
            throw new BuildException("The given file '" + dd + "' for the deployment descriptor does not exist.");
        }
        this.specificDeploymentDescriptor = dd;
    }

    /**
     * Sets the path to the archive that will be built.
     * @param dest the reference to resulting archive path.
     */
    public void setDest(final File dest) {
        this.dest = dest;
    }

    /**
     * Gets the path to the archive that will be built.
     * @return the reference to resulting archive path.
     */
    public File getDest() {
        return dest;
    }

    /**
     * Sets the relative Name of the archive (when embedded).
     * @param name the relative Name of the archive.
     */
    public void setName(final String name) {
        this.name = name;
    }

    /**
     * Gets the relative Name of the archive (when embedded).
     * @return the relative Name of the archive.
     */
    public String getName() {
        return name;
    }

    /**
     * Update the given archive info object with some attributes.
     * @param archiveInfo the object to update
     */
    protected void updateArchiveInfo(final ArchiveInfo archiveInfo) {
        archiveInfo.setDest(dest);
        archiveInfo.setDD(deploymentDescriptor);
        archiveInfo.setSpecificDD(specificDeploymentDescriptor);
        archiveInfo.setExploded(exploded);

        // Add the given fileset
        for (FileSet fileSet : getFileSetList()) {
            archiveInfo.addFileSet(fileSet);
        }

    }
}
