/**
 * EasyBeans
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsFactory.java 3037 2008-04-29 19:52:41Z sauthieg $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.container;

import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;

import javax.ejb.TimerService;
import javax.naming.Context;
import javax.naming.NamingException;

import org.ow2.easybeans.api.EZBContainer;
import org.ow2.easybeans.api.Factory;
import org.ow2.easybeans.api.FactoryException;
import org.ow2.easybeans.api.bean.EasyBeansBean;
import org.ow2.easybeans.api.components.EZBComponentRegistry;
import org.ow2.easybeans.api.injection.EasyBeansInjectionException;
import org.ow2.easybeans.api.injection.ResourceInjector;
import org.ow2.easybeans.component.itf.TimerComponent;
import org.ow2.easybeans.naming.NamingManager;
import org.ow2.easybeans.naming.interceptors.ENCManager;
import org.ow2.easybeans.rpc.api.EJBRequest;
import org.ow2.easybeans.rpc.api.EJBResponse;
import org.ow2.easybeans.rpc.util.Hash;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.Pool;
import org.ow2.util.pool.api.PoolException;

/**
 * Abstract factory which implements common and defaults methods.<br>
 * It should be extended by Bean factories.
 * @param <PoolType> the type of bean instance.
 * @author Florent Benoit
 */
public abstract class AbsFactory<PoolType extends EasyBeansBean> implements Factory<PoolType, Long> {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(AbsFactory.class);

    /**
     * Name of the class of the managed bean.
     */
    private String className = null;

    /**
     * Container that created this factory.
     */
    private EZBContainer container = null;

    /**
     * Pool that manage beans instance.
     */
    private Pool<PoolType, Long> pool = null;

    /**
     * Class used to build bean's instance.
     */
    private Class<PoolType> beanClass = null;

    /**
     * Context for java: lookups.
     */
    private Context javaContext = null;

    /**
     * Reference on the naming manager.
     */
    private static NamingManager namingManager = null;

    /**
     * List of external Resources injectors.
     */
    private List<ResourceInjector> injectors = null;

    /**
     * Keep a direct reference to the method so that we don't need to compute
     * each time the method object to invoke.<br>
     * http://java.sun.com/j2se/1.5.0/docs/guide/rmi/spec/rmi-stubs24.html
     */
    private Map<Long, Method> hashes = null;

    /**
     * Id of this container.
     */
    private String id = null;

    /**
     * Timer Service for this factory.
     */
    private TimerService timerService = null;


    /**
     * Builds a new factory with a given name and its container.
     * @param className name of this factory (name of class that is managed)
     * @param container the root component of this factory.
     * @throws FactoryException if class can't be loaded.
     */
    @SuppressWarnings("unchecked")
    public AbsFactory(final String className, final EZBContainer container) throws FactoryException {
        this.className = className;
        this.container = container;
        this.id = String.valueOf(System.identityHashCode(this));
        Class clazz = null;
        try {
            clazz = getContainer().getClassLoader().loadClass(getClassName());
        } catch (ClassNotFoundException e) {
            throw new FactoryException("Cannot load the class for class name '" + getClassName() + "'", e);
        }
        setBeanClass(clazz);
        setHashes(Hash.hashClass(clazz));

        try {
            namingManager = NamingManager.getInstance();
        } catch (NamingException e) {
            throw new FactoryException("Cannot get instance of the naming manager", e);
        }

        this.injectors = container.getConfiguration().getInjectors();

        // Get a timer service if a timer component is present.
        EZBComponentRegistry registry = container.getConfiguration().getEZBServer().getComponentManager().getComponentRegistry();
        List<TimerComponent> timerComponents = registry.getComponents(TimerComponent.class);

        // use the first one if there is at least once
        if (timerComponents.size() > 0) {
            TimerComponent timerComponent = timerComponents.get(0);
            if (timerComponents.size() > 1) {
                logger.warn("There are {0} timer components running on this server. Only the first one will be used",
                        new Integer(timerComponents.size()));
            }

            // Build a new timer service
            this.timerService = timerComponent.getTimerService(this);
        }

    }

    /**
     * Callback called when object is gonna be removed.
     * @param instance that is being removed from the pool.
     */
    public void remove(final PoolType instance) {
        ClassLoader oldClassLoader = Thread.currentThread().getContextClassLoader();
        Thread.currentThread().setContextClassLoader(getContainer().getClassLoader());
        try {
            // call callbacks
            instance.preDestroyEasyBeansLifeCycle();
        } catch (Exception e) {
            logger.error("Could not complete preDestroy method on instance", e);
        } finally {
            Thread.currentThread().setContextClassLoader(oldClassLoader);
        }

        // I need an extension point here for desinjection
        // call external resources injectors
        for (ResourceInjector injector : this.injectors) {
            try {
                injector.postEasyBeansDestroy(instance);
            } catch (Throwable t) {
                // Protection from malicious code
                logger.error("postEasyBeansDestroy() for {0} failed", injector.getClass().getName(), t);
            }
        }
    }

    /**
     * Injects Resources into the Bean.
     * @param instance The Bean instance to be injected.
     * @throws PoolException if resources cannot be injected.
     */
    protected void injectResources(final PoolType instance) throws PoolException {

        // call external resources injectors
        for (ResourceInjector injector : this.injectors) {
            try {
                injector.preEasyBeansInject(instance);
            } catch (Throwable t) {
                // Protection from malicious code
                logger.error("preEasyBeansInject() for {0} failed", injector.getClass().getName(), t);
            }
        }

        // call dependency injection
        try {
            instance.injectedByEasyBeans();
        } catch (EasyBeansInjectionException e) {
            throw new PoolException("Cannot inject resources in the created bean", e);
        }

        // call external resources injectors
        for (ResourceInjector injector : this.injectors) {
            try {
                injector.postEasyBeansInject(instance);
            } catch (Throwable t) {
                // Protection from malicious code
                logger.error("postEasyBeansInject() for {0} failed", injector.getClass().getName(), t);
            }
        }
    }

    /**
     * Gets the computed hashes.
     * @return computed hashes
     */
    protected Map<Long, Method> getHashes() {
        return this.hashes;
    }

    /**
     * Sets the hashes for the current bean class.
     * @param hashes method hashes computed as RMI hashes
     */
    protected void setHashes(final Map<Long, Method> hashes) {
        this.hashes = hashes;
    }

    /**
     * Gets the java: context.
     * @return java: context.
     */
    public Context getJavaContext() {
        return this.javaContext;
    }

    /**
     * Sets the java: context.
     * @param javaContext the java: context.
     */
    public void setJavaContext(final Context javaContext) {
        // Can be only set once
        if (this.javaContext != null) {
            throw new IllegalStateException("The javaContext can only be set once. Already set !");
        }
        this.javaContext = javaContext;

        // Set the javaContext used for ENC
        ENCManager.initContext(this, javaContext);
    }

    /**
     * Gets the bean's class.
     * @return bean class used to instantiate beans.
     */
    public Class<PoolType> getBeanClass() {
        return this.beanClass;
    }

    /**
     * Sets the bean class that will be used to build bean's instance.
     * @param beanClass the instance of the bean class name
     */
    protected void setBeanClass(final Class<PoolType> beanClass) {
        this.beanClass = beanClass;
    }

    /**
     * Sets the pool used by this factory.
     * @param pool the pool which managed bean instances
     */
    protected void setPool(final Pool<PoolType, Long> pool) {
        this.pool = pool;
    }

    /**
     * Gets the container used by this factory.
     * @return container of this factory
     */
    public EZBContainer getContainer() {
        return this.container;
    }

    /**
     * Gets the className used by this factory.
     * @return classname that will be instantiated to build bean instance.
     */
    public String getClassName() {
        return this.className;
    }

    /**
     * Gets the reference on the naming manager.
     * @return the reference on the naming manager.
     */
    protected static NamingManager getNamingManager() {
        return namingManager;
    }


    /**
     * Gets the pool used by this factory.
     * @return pool.
     */
    public Pool<PoolType, Long> getPool() {
        return this.pool;
    }

    /**
     * A request comes to the bean factory and needs to be handled.<br>
     * A response is done which contains the answer.
     * @param request the EJB request.
     * @return a response that have been processed by the factory.
     */
    public abstract EJBResponse rpcInvoke(final EJBRequest request);


    /**
     * Init the factory.
     * @throws FactoryException if the initialization fails.
     */
    public void init() throws FactoryException {

    }


    /**
     * Gets the id of this container.
     * @return string id.
     */
    public String getId() {
        return this.id;
    }

    /**
     * Stops the factory.
     */
    public void stop() {
        ENCManager.removeContext(this);
    }


    /**
     * Gets the timer service of this factory.
     * @return the timer service.
     */
    public TimerService getTimerService() {
        return this.timerService;
    }

}
