/**
 * EasyBeans
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ContainersMonitor.java 2842 2008-04-11 15:22:31Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.server;

import static org.ow2.easybeans.util.url.URLUtils.urlToFile;

import java.io.File;
import java.net.URL;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import org.ow2.easybeans.api.EZBContainer;
import org.ow2.easybeans.api.EZBContainerException;
import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class monitors the archives managed by containers and reload them if the
 * archive is changed.
 * @author Florent Benoit
 */
public class ContainersMonitor extends Thread {

    /**
     * Sleep time for the thread of the cleaner (5s).
     */
    private static final int SLEEP_TIME = 5000;

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(ContainersMonitor.class);

    /**
     * Map between container and the last updated file.
     */
    private Map<EZBContainer, Long> modifiedFiles = null;

    /**
     * Embedded server which is monitored.
     */
    private Embedded embedded = null;

    /**
     * Initializing period ?.
     */
    private boolean bootInProgress = false;

    /**
     * Stop order received ?
     */
    private boolean stopped = false;

    /**
     * Builds a new monitor by initializing lists.
     * @param embedded the embedded server which is monitored.
     */
    public ContainersMonitor(final Embedded embedded) {
        this.embedded = embedded;
        this.modifiedFiles = new WeakHashMap<EZBContainer, Long>();
    }


    /**
     * Init containers (call at startup).
     */
    public void init() {
        // Start the boot process.
        this.bootInProgress = true;

        // Initialize containers
        scanNewContainers();

        // No more in the boot process.
        this.bootInProgress = false;
    }

    /**
     * Start the thread of this class It will clean all the work entries.
     */
    @Override
    public void run() {

        for (;;) {
            // Stop the thread
            if (this.stopped) {
                return;
            }


            // Check existing container is not modified ?
            for (EZBContainer container : this.embedded.getContainers().values()) {
                if (container.isAvailable()) {
                    checkContainer(container);
                }
            }

            // Check new containers to start
            scanNewContainers();

            try {
                Thread.sleep(SLEEP_TIME);
            } catch (InterruptedException e) {
                throw new RuntimeException("Thread fail to sleep");
            }
        }
    }

    /**
     * Scan all files present in the ejb3 directory and create container for
     * each one.
     */
    private void scanNewContainers() {

        // Get the list of deploy directories
        List<File> deployDirectories = this.embedded.getServerConfig().getDeployDirectories();

        for (File deployDirectory : deployDirectories) {
            // get files
            File[] files = deployDirectory.listFiles();

            // next directory if there are no files to scan.
            if (files == null) {
                continue;
            }

            // analyze each file.
            for (File f : files) {
                if (f.getName().toLowerCase().endsWith(".jar")) {
                    IArchive archive = ArchiveManager.getInstance().getArchive(f);

                    // check that no container already exist for this archive
                    boolean alreadyExist = false;
                    for (EZBContainer container : this.embedded.getContainers().values()) {
                        if (container.getArchive().equals(archive)) {
                            alreadyExist = true;
                        }

                    }
                    if (!alreadyExist) {
                        if (!this.bootInProgress) {
                            // wait that files are copied before creating/starting
                            // the container
                            try {
                                Thread.sleep(SLEEP_TIME);
                            } catch (InterruptedException e) {
                                throw new RuntimeException("Thread fail to sleep");
                            }
                        }
                        this.logger.info("Creating container for archive {0}.", f);

                        EZBContainer container = this.embedded.createContainer(archive);
                        try {
                            container.start();
                        } catch (EZBContainerException e) {
                            this.logger.error("Cannot start container {0}", container.getName(), e);
                        }
                    }
                }
            }
        }
    }

    /**
     * Check a container (and its archive) and see if there is a need to reload
     * the container.
     * @param container the container to monitor.
     */
    protected void checkContainer(final EZBContainer container) {
        // get lastmodified for this container (any ?)
        long previousLastModified = 0;
        Long l = this.modifiedFiles.get(container);
        if (l != null) {
            previousLastModified = l.longValue();
        }

        // get archive
        IArchive archive = container.getArchive();

        // Get URL
        URL url = null;
        try {
            url = archive.getURL();
        } catch (ArchiveException e1) {
            this.logger.warn("Cannot get URL on the container {0}", archive.getName());
            return;
        }
        File file = urlToFile(url);
        // No file archive, abort.
        if (!file.exists()) {
            return;
        }

        long updatedModified = getLastModified(file);
        this.modifiedFiles.put(container, Long.valueOf(updatedModified));

        // first check. nothing to do
        if (previousLastModified == 0) {
            return;
        }
        // container was modified, need to relaunch it
        if (updatedModified > previousLastModified) {
            this.logger.info("Container with archive {0} was modified. Reloading...", archive.getName());
            container.stop();
            try {
                container.start();
            } catch (EZBContainerException e) {
                e.printStackTrace();
            }
        }

    }

    /**
     * Gets the last modified attribute of a given archive.<br> If it is a
     * directory, returns the last modified file of the archive.
     * @param archive the archive to monitor.
     * @return the last modified version of the given archive.
     */
    protected long getLastModified(final File archive) {
        if (archive.isFile()) {
            return archive.lastModified();
        }
        // else
        File[] files = archive.listFiles();
        long last = 0;
        if (files != null) {
            for (File f : files) {
                last = Math.max(last, getLastModified(f));
            }
        }
        return last;
    }


    /**
     * Receives a stop order.
     */
    public void stopOrder() {
        this.stopped = true;
    }

    /**
     * Gets the embedded object.
     * @return embedded object
     */
    public Embedded getEmbedded() {
        return this.embedded;
    }
}
